/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 *
 * Generated by generate_moe_gemm_tma_ws_sm120_fp4.py. Do not edit manually.
 */

#ifndef EXCLUDE_SM_120
#ifdef COMPILE_BLACKWELL_SM120_TMA_GROUPED_GEMMS
#if defined(ENABLE_FP4) && defined(USE_FP4_QMOE)

#include "contrib_ops/cuda/llm/moe_gemm/launchers/moe_gemm_tma_ws_launcher.inl"

namespace onnxruntime::llm::kernels::cutlass_kernels {

INSTANTIATE_TMA_WARP_SPECIALIZED_MOE_GEMM(Sm120, SafeFP4, SafeFP4, half, EpilogueOpDefault, NONE, 256, 128, 128, 1, 1, 1, false, false)

}  // namespace onnxruntime::llm::kernels::cutlass_kernels

#endif  // ENABLE_FP4 && USE_FP4_QMOE
#endif  // COMPILE_BLACKWELL_SM120_TMA_GROUPED_GEMMS
#endif  // EXCLUDE_SM_120
