// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#define HEAD_ELEMS 128
#define HEAD_DIM_NAMESPACE H128

#ifdef USE_FP8_KV_CACHE
#include "xqa_loader_bf16_fp8_impl.cuh"
#endif
